// testcrange.cpp

// incrementer, at_end and deref functions are kept internally, no virtual calls
namespace func_calls {
    template< class type, class container> struct crange_impl;

    template< class type> struct crange {
    protected:
        typedef crange<type> self_type;
        typedef void (*incrementer)(const self_type&);
        typedef bool (*at_end)(const self_type &);
        typedef type & (*deref)(const self_type&);

        crange( incrementer inc_func, at_end at_end_func, deref deref_func) 
            : m_inc_func(inc_func), m_at_end_func(at_end_func), m_deref_func(deref_func) {}
    public:

        operator bool() const                   { return !m_at_end_func(*this); }
        const self_type & operator++() const    { m_inc_func(*this); return *this; }
        type & operator*() const                { return m_deref_func(*this); }
    private:
        const incrementer m_inc_func;
        const at_end m_at_end_func;
        const deref m_deref_func;
    };

    template< class type, class container> struct crange_impl : public crange<type> {
        typedef crange<type> base;
        typedef crange_impl<type,container> self_type;
        crange_impl( container & c) : m_first(c.begin()), m_last(c.end()), 
            base(&self_type::increment, &self_type::at_end, &self_type::deref) {}

    private:
        static void increment(const base & me) {
            const self_type & self = reinterpret_cast<const self_type&>(me);
            ++self.m_first;
        }
        static bool at_end(const base & me) {
            const self_type & self = reinterpret_cast<const self_type&>(me);
            return self.m_first == self.m_last;
        }
        static type & deref(const base & me) {
            const self_type & self = reinterpret_cast<const self_type&>(me);
            return *self.m_first;
        }
    private:
        mutable typename container::iterator m_first, m_last;
    };


    template<class container> crange_impl<typename container::value_type, container> mkrange(container & c) {
        typedef crange_impl<typename container::value_type, container> impl;
        return impl(c);
    }
} // namespace internal_functions



// incrementer, at_end and deref functions are ***virtual*** calls
namespace vtable_calls {
    template< class type> struct crange_impl_base {
        virtual ~crange_impl_base() {}
        virtual operator bool() const = 0;
        virtual void operator++() = 0;
        virtual type & operator*() = 0;
    };

    template< class type, class container> struct crange_impl : public crange_impl_base<type> {
        crange_impl( container & c) : m_first(c.begin()), m_last(c.end()) {}

        operator bool() const   { return m_first != m_last; }
        void operator++()       { ++m_first; }
        type & operator*()      { return *m_first; }
    private:
        typename container::iterator m_first, m_last;
    };

    template< class type> struct crange {
    private:
        crange( const crange<type> & other) {} // note: most likely, should not be implemented
    public:
        template<class container> crange( container & c) : impl( new crange_impl<type,container>(c)) {}
        ~crange() { delete impl; }

        operator bool() const       { return impl->operator bool(); }
        crange<type> & operator++()    { ++*impl; return *this; }
        type & operator*()          { return impl->operator*(); }
    private:
        crange_impl_base<type> * impl;
    };
} // namespace vtable_calls 


#include "foreach.hpp" // Eric Niebler's foreach

#include <vector>
#include <list>
#include <deque>
#include <algorithm>
#include <stdlib.h>
#include <sstream>
#include <time.h>
#include <fstream>

// contains a number, as a string
struct str_sequence {
    str_sequence( int i) { set(i); }
    void set( int i) {
        std::ostringstream out;
        out << i;
        str = out.str();
    }
    std::string str;
};



////////////////////////////////////////////////////////////////////////////////////////////////
// Test functions

// ... note: they all modify the value, somehow. calling this function twice with the same reference,
//     will modify the value twice (unlike something like void setzero(int & val) { val = 0; }

struct set_to_idx {
    set_to_idx() : idx(0) {}
    template< class T> void operator()(T & val) { val = idx++; }
private:
    int idx;
};

inline void multiply_by_2(int & i)                 { i *= 2; }
inline void multiply_by_max20(int & i)             { i *= rand() % 20; }
inline void add_one_char( str_sequence & s)        { s.str.resize( s.str.size() + 1); }
inline void add_10(int & i)                        { i += 10; } 
inline void noop_int( int & )                      {}
inline void noop_seq( str_sequence & )             {}
inline void set_to_len( str_sequence & s)          { 
    s.set( s.str.length() + 1); 
    s.str.resize( s.str.length() * 5, s.str[0]); 
}

struct set_to_az {
    set_to_az() : end('a') {}
    void operator()( str_sequence & s) {
        s.str.erase();
        for ( char start = 'a'; start < end; ++start)
            s.str = start + s.str;
        if ( ++end == 'z') end = 'a';
    }
private:
    char end;
};


////////////////////////////////////////////////////////////////////////////////////////////////
// Runs the tests

// modify these to suit your needs (also, you can use command-line arguments)
int MAX_ELEMS_COUNT = 128000; // how many elements to add in each array?
int ELEMS_COUNT = 128000; // how many elements to add in each array? (is doubled, until reaches MAX_ELEMS_COUNT)
const int TESTS = 3; // how many tests

bool first_test_run = false;
int run_times_so_far = 0;
double sum_raw = 0;
double sum_func = 0;
double sum_vtable = 0;
double sum_each = 0;
double sum_func_percentage = 0;
double sum_vtable_percentage = 0;
double sum_each_percentage = 0;

std::ofstream results_file( "result.txt");

template<class container>
void fill_container( container & c, int elems) {
    for ( int idx = 0; idx < elems; ++idx) 
        c.push_back( rand() % 10000); // not too big numbers, we don't want overflows
}

template<class container, class test_func>
inline void run_test(const std::string & test_name, const container & , test_func f, int times, int elems) {
    container copy;
    fill_container(copy, elems);
    container c; // run all tests on exactly the same content of the container

    // use raw iterators
    c = copy; // run all tests on exactly the same content of the container
    clock_t start = clock();
    test_func call = f;
    for ( int idx = 0; idx < times; ++idx)
        for ( container::iterator b = c.begin(), e = c.end(); b != e; ++b)
            call(*b);
    clock_t end = clock();
    double raw_secs = (double)(end - start) / CLOCKS_PER_SEC;

    // use func_calls iterators
    c = copy; // run all tests on exactly the same content of the container
    typedef container::value_type value;
    start = clock();
    call = f;
    { using namespace func_calls;
      for ( int idx = 0; idx < times; ++idx)
        for ( crange<value> const & r = mkrange(c); r; ++r)
            call(*r); }
    end = clock();
    double func_secs = (double)(end - start) / CLOCKS_PER_SEC;
    
    // use vtable_calls iterators
    c = copy; // run all tests on exactly the same content of the container
    typedef container::value_type value;
    start = clock();
    call = f;
    { using namespace vtable_calls;
      for ( int idx = 0; idx < times; ++idx)
        for ( crange<value> r(c); r; ++r)
            call(*r); }
    end = clock();
    double vtable_secs = (double)(end - start) / CLOCKS_PER_SEC;

    // use EricNiebler's for_each
    c = copy; // run all tests on exactly the same content of the container
    typedef container::value_type value;
    start = clock();
    call = f;
    { using namespace vtable_calls;
      for ( int idx = 0; idx < times; ++idx)
        BOOST_FOREACH( value & v, c)
            call(v); }
    end = clock();
    double each_secs = (double)(end - start) / CLOCKS_PER_SEC;
    
    // john.torjo - added after I've come up with the results
    // ... account for innacturate clock
    if ( func_secs < raw_secs) func_secs = raw_secs;
    if ( vtable_secs < raw_secs) vtable_secs = raw_secs;
    if ( each_secs < raw_secs) each_secs = raw_secs;

    std::ostringstream msg;
    msg << "\nTest " << test_name << " took "
        << "\n  RAW            : " << raw_secs << " secs"
        << "\n  func calls     : " << func_secs << " secs";
    if ( raw_secs > 0)
        msg << " (" << ((func_secs * 100) / raw_secs) << "%)";
    msg << "\n  vtable calls: " << vtable_secs << " secs";
    if ( raw_secs > 0)
        msg << " (" << ((vtable_secs * 100) / raw_secs) << "%)";
    msg << "\n  foreach calls: " << each_secs << " secs";
    if ( raw_secs > 0)
        msg << " (" << ((each_secs * 100) / raw_secs) << "%)";
    results_file << msg.str() << std::endl;
    std::cout << msg.str() << std::endl;

    // when computing averages, we're ignoring first test (we're just warming up ;))
    if ( first_test_run)
        if ( raw_secs > 0) {
            ++run_times_so_far;
            sum_raw += raw_secs;
            sum_func += func_secs;
            sum_vtable += vtable_secs;
            sum_each += each_secs;
            sum_func_percentage += ((func_secs * 100) / raw_secs);
            sum_vtable_percentage += ((vtable_secs * 100) / raw_secs);
            sum_each_percentage += ((each_secs * 100) / raw_secs);
        }
}


void test() {
    typedef std::vector<int> int_vector;
    typedef std::deque<int> int_deque;
    typedef std::list<int> int_list;

    typedef std::vector<str_sequence> seq_vector;
    typedef std::deque<str_sequence> seq_deque;
    typedef std::list<str_sequence> seq_list;

    int_vector vi;
    int_deque di;
    int_list li;
    seq_vector vs;
    seq_deque ds;
    seq_list ls;

    srand( (unsigned)time(0));

    const int TIMES = 10; // how many times to run each test?
    run_test( "set_to_idx, vector<int>", vi,       set_to_idx(), TIMES * 10, ELEMS_COUNT * 10);
    run_test( "set_to_idx, list<int>", li,         set_to_idx(), TIMES * 10, ELEMS_COUNT * 10);
    run_test( "set_to_idx, deque<int>", di,        set_to_idx(), TIMES * 10, ELEMS_COUNT * 10);
    run_test( "set_to_idx, vector<seq>", vs,       set_to_idx(), TIMES, ELEMS_COUNT);
    run_test( "set_to_idx, list<seq>", ls,         set_to_idx(), TIMES, ELEMS_COUNT);
    run_test( "set_to_idx, deque<seq>", ds,        set_to_idx(), TIMES, ELEMS_COUNT);

    run_test( "multiply_by_2, vector<int>", vi,       multiply_by_2, TIMES * 10, ELEMS_COUNT * 10);
    run_test( "multiply_by_2, list<int>", li,         multiply_by_2, TIMES * 10, ELEMS_COUNT * 10);
    run_test( "multiply_by_2, deque<int>", di,        multiply_by_2, TIMES * 10, ELEMS_COUNT * 10);

    run_test( "multiply_by_max20, vector<int>", vi,       multiply_by_max20, TIMES * 10, ELEMS_COUNT * 20);
    run_test( "multiply_by_max20, list<int>", li,         multiply_by_max20, TIMES * 10, ELEMS_COUNT * 15);
    run_test( "multiply_by_max20, deque<int>", di,        multiply_by_max20, TIMES * 10, ELEMS_COUNT * 15);

    run_test( "add_10, vector<int>", vi,       add_10, TIMES * 5, ELEMS_COUNT * 10);
    run_test( "add_10, list<int>", li,         add_10, TIMES * 5, ELEMS_COUNT * 10);
    run_test( "add_10, deque<int>", di,        add_10, TIMES * 5, ELEMS_COUNT * 10);

    run_test( "add_one_char, vector<seq>", vs,       add_one_char, TIMES * 5, ELEMS_COUNT);
    run_test( "add_one_char, list<seq>", ls,         add_one_char, TIMES * 5, ELEMS_COUNT);
    run_test( "add_one_char, deque<seq>", ds,        add_one_char, TIMES * 5, ELEMS_COUNT);

    run_test( "set_to_len, vector<seq>", vs,       set_to_len, TIMES, ELEMS_COUNT);
    run_test( "set_to_len, list<seq>", ls,         set_to_len, TIMES, ELEMS_COUNT);
    run_test( "set_to_len, deque<seq>", ds,        set_to_len, TIMES, ELEMS_COUNT);

    run_test( "set_to_az, vector<seq>", vs,       set_to_az(), TIMES, ELEMS_COUNT);
    run_test( "set_to_az, list<seq>", ls,         set_to_az(), TIMES, ELEMS_COUNT);
    run_test( "set_to_az, deque<seq>", ds,        set_to_az(), TIMES, ELEMS_COUNT);

    run_test( "noop, vector<int>", vi,       noop_int, TIMES * 5, ELEMS_COUNT * 25);
    run_test( "noop, list<int>", li,         noop_int, TIMES * 5, ELEMS_COUNT * 25);
    run_test( "noop, deque<int>", di,        noop_int, TIMES * 5, ELEMS_COUNT * 25);
    run_test( "noop, vector<seq>", vs,       noop_seq, TIMES * 50, ELEMS_COUNT);
    run_test( "noop, list<seq>", ls,         noop_seq, TIMES * 50, ELEMS_COUNT);
    run_test( "noop, deque<seq>", ds,        noop_seq, TIMES * 50, ELEMS_COUNT);

}

std::stringstream avg;
void test_for_elems() {
    first_test_run = false;
    run_times_so_far = 0;
    sum_raw = 0;
    sum_func = 0;
    sum_vtable = 0;
    sum_each = 0;
    sum_func_percentage = 0;
    sum_vtable_percentage = 0;
    sum_each_percentage = 0;

    for ( int idx = 0; idx < TESTS; ++idx) {
        results_file << "Test " << idx + 1 << " of " << TESTS << " (for " << ELEMS_COUNT << " elements)" << std::endl;
        std::cout << "Test " << idx + 1 << " of " << TESTS << " (for " << ELEMS_COUNT << " elements)" << std::endl;
        test();
        first_test_run = true;
    }

    avg << "\nAverages for " << ELEMS_COUNT << "elements (except first test): \n" 
        << "raw time    : " << sum_raw << " secs \n"
        << "func calls  : " << sum_func << " secs (avg percentage " << sum_func_percentage / run_times_so_far << ")\n"
        << "vtable calls: " << sum_vtable << " secs (avg percentage " << sum_vtable_percentage / run_times_so_far << ")\n"
        << "foreach calls: " << sum_each << " secs (avg percentage " << sum_each_percentage / run_times_so_far << ")" << std::endl;
}

// results are outputted to "result.txt" also
int main(int argc, char * argv[]) {
    // Example of usage: 'testcrange 32000 128000'
    //                    will run each test for with 32000, 64000, 128000 elements
    if ( argc > 1) ELEMS_COUNT = atoi(argv[1]);
    if ( argc > 2) MAX_ELEMS_COUNT = atoi(argv[2]);

    clock_t start = clock();
    for ( int count = ELEMS_COUNT; count <= MAX_ELEMS_COUNT; count *= 2) {
        ELEMS_COUNT = count;
        test_for_elems();
    }
    clock_t end = clock();
    double total = (double)(end - start) / CLOCKS_PER_SEC;
    std::cout << avg.str() << std::endl;
    results_file << avg.str() << std::endl;
    std::cout << "Test took " << total << " secs " << std::endl;
    results_file << "Test took " << total << " secs " << std::endl;
	return 0;
}

